"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Technical analysis - basic metrics (:mod:`pynance.tech.simple`)
==================================================================

.. currentmodule:: pynance.tech.simple
"""

from __future__ import absolute_import

import numpy as np
import pandas as pd

def growth(eqdata, **kwargs):
    """
    Generate a DataFrame where the sole column, 'Growth',
    is the growth for the equity over the given number of sessions.
    
    For example, if 'XYZ' has 'Adj Close' of `100.0` on 2014-12-15 and 
    `90.0` 4 *sessions* later on 2014-12-19, then the 'Growth' value
    for 2014-12-19 will be `0.9`.

    Parameters
    ----------
    eqdata : DataFrame
        Data such as that returned by :func:`pynance.data.retrieve.get`
    selection : str, optional
        Column from which to determine growth values. Defaults to
        'Adj Close'.
    n_sessions : int
        Number of sessions to count back for calculating today's
        growth. For example, if `n_sessions` is set to 4, growth is
        calculated relative to the price 4 sessions ago. Defaults
        to 1 (price of previous session).
    skipstartrows : int
        Rows to skip at beginning of `eqdata` in addition to the 1 row that must
        be skipped because the calculation relies on a prior data point.
        Defaults to 0.
    skipendrows : int
        Rows to skip at end of `eqdata`. Defaults to 0.
    outputcol : str, optional
        Name to use for output column. Defaults to 'Growth'

    Returns
    ----------
    out : DataFrame

    Notes
    ----------
    The interval is the number of *sessions* between the 2 values
    whose ratio is being measured, *not* the number of days (which
    includes days on which the market is closed).

    Growth is measured relative to the earlier
    date, but the index date is the later date. This index is chosen because
    it is the date on which the value is known.
    """
    selection = kwargs.get('selection', 'Adj Close')
    n_sessions = kwargs.get('n_sessions', 1)
    skipstartrows = kwargs.get('skipstartrows', 0)
    skipendrows = kwargs.get('skipendrows', 0)
    outputcol = kwargs.get('outputcol', 'Growth')
    size = len(eqdata.index)
    growthdata = eqdata.loc[:, selection].values[(skipstartrows + n_sessions):(size - skipendrows)] / \
            eqdata.loc[:, selection].values[skipstartrows:(-n_sessions - skipendrows)]
    growthindex = eqdata.index[(skipstartrows + n_sessions):(size - skipendrows)]
    return pd.DataFrame(data=growthdata, index=growthindex, columns=[outputcol], dtype='float64')

def ln_growth(eqdata, **kwargs):
    """
    Return the natural log of growth.

    See also
    --------
    :func:`growth`
    """
    if 'outputcol' not in kwargs:
        kwargs['outputcol'] = 'LnGrowth'
    return np.log(growth(eqdata, **kwargs))

def ret(eqdata, **kwargs):
    """
    Generate a DataFrame where the sole column, 'Return',
    is the return for the equity over the given number of sessions.
    
    For example, if 'XYZ' has 'Adj Close' of `100.0` on 2014-12-15 and 
    `90.0` 4 *sessions* later on 2014-12-19, then the 'Return' value
    for 2014-12-19 will be `-0.1`.

    Parameters
    ----------
    eqdata : DataFrame
        Data such as that returned by `get()`
    selection : str, optional
        Column from which to determine growth values. Defaults to
        'Adj Close'.
    n_sessions : int
        Number of sessions to count back for calculating today's
        return. For example, if `n_sessions` is set to 4, return is
        calculated relative to the price 4 sessions ago. Defaults
        to 1 (price of previous session).
    skipstartrows : int
        Rows to skip at beginning of `eqdata` in addition to the 1 row that must
        be skipped because the calculation relies on a prior data point.
        Defaults to 0.
    skipendrows : int
        Rows to skip at end of `eqdata`. Defaults to 0.
    outputcol : str, optional
        Name for column of output dataframe. Defaults to 'Return'.

    Returns
    ----------
    out : DataFrame

    Notes
    ----------
    The interval is the number of *sessions* between the 2 values
    whose ratio is being measured, *not* the number of days (which
    includes days on which the market is closed).

    The percentage gain or loss is measured relative to the earlier
    date, but the index date is the later date. The index is chose because
    that is the date on which the value is known. The percentage measure is because
    that is the way for calculating percent profit and loss.
    """
    if 'outputcol' not in kwargs:
        kwargs['outputcol'] = 'Return'
    result = growth(eqdata, **kwargs)
    result.values[:, :] -= 1.
    return result



"""
.. Copyright (c) 2014 Marshall Farrier
   license http://opensource.org/licenses/MIT

Data - building features (:mod:`pynance.data.feat`)
==============================================================

.. currentmodule:: pynance.data.feat

These functions are intended to be used in conjunction
with `functools.partial` and other function decorators
to pass to `data.labeledfeatures()`.
For example,

>>> from functools import partial
>>> featfunc = pn.decorate(partial(pn.data.feat.fromfuncs, [fn1, fn2, fn3], skipatstart=averaging_window), 
        averaging_window + n_feature_sessions - 1)
>>> features, labels = pn.data.labeledfeatures(eqdata, featfunc, labelfunc) 
"""

from functools import partial

import numpy as np
import pandas as pd

def add_const(features):
    """
    Prepend the constant feature 1 as first feature and return the modified
    feature set.

    Parameters
    ----------
    features : ndarray or DataFrame
    """
    content = np.empty((features.shape[0], features.shape[1] + 1), dtype='float64')
    content[:, 0] = 1.
    if isinstance(features, np.ndarray):
        content[:, 1:] = features
        return content
    content[:, 1:] = features.iloc[:, :].values
    cols = ['Constant'] + features.columns.tolist()
    return pd.DataFrame(data=content, index=features.index, columns=cols, dtype='float64')

def fromcols(selection, n_sessions, eqdata, **kwargs):
    """
    Generate features from selected columns of a dataframe.

    Parameters
    ----------
    selection : list or tuple {str}
        Columns to be used as features.

    n_sessions : int
        Number of sessions over which to create features.

    eqdata : DataFrame
        Data from which to generate feature set. Must contain
        as columns the values from which the features are to
        be generated.

    constfeat : bool, optional
        Whether or not the returned features will have the constant
        feature.

    Returns
    ----------
    features : DataFrame
    """
    _constfeat = kwargs.get('constfeat', True)
    _outcols = ['Constant'] if _constfeat else []
    _n_rows = len(eqdata.index)
    for _col in selection:
        _outcols += map(partial(_concat, strval=' ' + _col), range(-n_sessions + 1, 1))
    _features = pd.DataFrame(index=eqdata.index[n_sessions - 1:], columns=_outcols, dtype=np.float64)
    _offset = 0
    if _constfeat:
        _features.iloc[:, 0] = 1.
        _offset += 1
    for _col in selection:
        _values = eqdata.loc[:, _col].values
        for i in range(n_sessions):
            _features.iloc[:, _offset + i] = _values[i:_n_rows - n_sessions + i + 1]
        _offset += n_sessions
    return _features

def fromfuncs(funcs, n_sessions, eqdata, **kwargs):
    """
    Generate features using a list of functions to apply to input data

    Parameters
    ----------
    funcs : list {function}
        Functions to apply to eqdata. Each function is expected
        to output a dataframe with index identical to a slice of `eqdata`.
        The slice must include at least `eqdata.index[skipatstart + n_sessions - 1:]`.
        Each function is also expected to have a function attribute
        `title`, which is used to generate the column names of the
        output features.

    n_sessions : int
        Number of sessions over which to create features.

    eqdata : DataFrame
        Data from which to generate features. The data will often
        be retrieved using `pn.get()`.

    constfeat : bool, optional
        Whether or not the returned features will have the constant
        feature.

    skipatstart : int, optional
        Number of rows to omit at the start of the output DataFrame.
        This parameter is necessary if any of the functions requires
        a rampup period before returning valid results, e.g. `sma()` or
        functions calculating volume relative to a past baseline.
        Defaults to 0.

    Returns
    ----------
    features : DataFrame
    """
    _skipatstart = kwargs.get('skipatstart', 0)
    _constfeat = kwargs.get('constfeat', True)
    _outcols = ['Constant'] if _constfeat else []
    _n_allrows = len(eqdata.index)
    _n_featrows = _n_allrows - _skipatstart - n_sessions + 1
    for _func in funcs:
        _outcols += map(partial(_concat, strval=' ' + _func.title), range(-n_sessions + 1, 1))
    _features = pd.DataFrame(index=eqdata.index[_skipatstart + n_sessions - 1:],
            columns=_outcols, dtype=np.float64)
    _offset = 0
    if _constfeat:
        _features.iloc[:, 0] = 1.
        _offset += 1
    for _func in funcs:
        _values = _func(eqdata).values
        _n_values = len(_values)
        for i in range(n_sessions):
            _val_end = _n_values - n_sessions + i + 1
            _features.iloc[:, _offset + i] = _values[_val_end - _n_featrows:_val_end]
        _offset += n_sessions
    return _features

def _concat(intval, strval):
    return str(intval) + strval



"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Data - building labels (:mod:`pynance.data.lab`)
====================================================

.. currentmodule:: pynance.data.lab

These functions are intended to be used in conjunction
with :func:`functools.partial` to pass to 
:func:`pynance.data.combine.labeledfeatures`.
For example::

    >>> from functools import partial
    >>> features, labels = pn.data.labeledfeatures(eqdata, 256,
    ...        partial(pn.data.lab.growth, 32))
"""

import pandas as pd

def growth(interval, pricecol, eqdata):
    """
    Retrieve growth labels.

    Parameters
    --------------
    interval : int
        Number of sessions over which growth is measured. For example, if
        the value of 32 is passed for `interval`, the data returned will 
        show the growth 32 sessions ahead for each data point.
    eqdata : DataFrame
        Data for evaluating growth.
    pricecol : str
        Column of `eqdata` to be used for prices (Normally 'Adj Close').

    Returns
    --------
    labels : DataFrame
        Growth labels for the specified period
    skipatend : int
        Number of rows skipped at the end of `eqdata` for the given labels.
        Used to synchronize labels and features.

    Examples
    ---------------
    >>> from functools import partial
    >>> features, labels = pn.data.labeledfeatures(eqdata, 256, 
    ...        partial(pn.data.lab.growth, 32, 'Adj Close'))
    """
    size = len(eqdata.index)
    labeldata = eqdata.loc[:, pricecol].values[interval:] /\
            eqdata.loc[:, pricecol].values[:(size - interval)]
    df = pd.DataFrame(data=labeldata, index=eqdata.index[:(size - interval)],
            columns=['Growth'], dtype='float64')
    return df



"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Data - preprocessing functions (:mod:`pynance.data.prep`)
=========================================================

.. currentmodule:: pynance.data.prep
"""

import numpy as np
import pandas as pd

def center(dataset, out=None):
    """
    Returns a centered data set.
    
    Each column of the returned data will have mean 0.
    The row vector subtracted from each row to achieve this
    transformation is also returned.

    Parameters
    ----------
    dataset : DataFrame or ndarray

    out : DataFrame or ndarray, optional
        Alternate output array in which to place the result.
        If provided, it must have the same shape and type
        (DataFrame or ndarray) as the expected output.

    Returns
    ----------
    out : tuple of DataFrame or ndarray
        The output data is of the same type as the input.

    Notes
    ----------
    To exclude a column (such as a constant feature, which is
    usually the first or last column of data) simply don't
    include it in the input. For example:

    >>> centered_data, means = pn.center(mydata.iloc[:, 1:])

    To perform this operation in place:

    >>> _, means = pn.center(mydata.iloc[:, 1:], out=mydata.iloc:, 1:])
    """
    return _preprocess(_center_fn, dataset, out)

def _preprocess(func, dataset, out):
    # Generic preprocessing function used in center() and normalize()
    is_df = isinstance(dataset, pd.DataFrame)
    _data = (dataset.values if is_df else dataset)
    processed_data, adjustment = func(_data)
    if not is_df:
        if out is not None:
            out[:, :] = processed_data
            return out, adjustment
        return processed_data, adjustment
    adj_df = pd.DataFrame(data=adjustment, index=['Mean'], columns=dataset.columns,
            dtype='float64')
    if out is not None:
        out.values[:, :] = processed_data
        return out, adj_df
    processed_df = pd.DataFrame(data=processed_data, index=dataset.index, 
            columns=dataset.columns, dtype='float64')
    return processed_df, adj_df

def _center_fn(_data):
    adjustment = np.mean(_data, axis=0, dtype=np.float64).reshape((1, _data.shape[1]))
    centered_data = _data - adjustment
    return centered_data, adjustment

def _normalize_fn(_data):
    adjustment = np.std(_data, axis=0, dtype=np.float64).reshape((1, _data.shape[1]))
    normalized_data = _data / adjustment
    return normalized_data, adjustment

def normalize(centered_data, out=None):
    """
    Returns a data set with standard deviation of 1.

    The input data must be centered for the operation to
    yield valid results: The mean of each column must be 0.
    Each column of the returned data set will have standard
    deviation 1.

    The row vector by which each row of data is divided is
    also returned.

    Parameters
    ----------
    centered_data : DataFrame or ndarray

    out : DataFrame or ndarray, optional
        Alternate output array in which to place the result.
        If provided, it must have the same shape and type
        (DataFrame or ndarray) as the expected output.

    Returns
    ----------
    out : tuple of DataFrame or ndarray
        The output data is of the same type as the input.

    Notes
    ----------
    To exclude a column (such as a constant feature, which is
    usually the first or last column of data) simply don't
    include it in the input. For example:

    >>> normalized_data, sd_adj = pn.normalize(mydata.iloc[:, 1:])

    To perform this operation in place:

    >>> _, sd_adj = pn.normalize(mydata.iloc[:, 1:], out=mydata.iloc:, 1:])
    """
    return _preprocess(_normalize_fn, centered_data, out)

def transform(data_frame, **kwargs):
    """
    Return a transformed DataFrame

    Transform data_frame along the given axis. By default, each row
    will be normalized (axis=0)

    Parameters
    -----------
    data_frame : DataFrame
        data to be normalized

    axis : int in {0, 1}, default: 0
        0 to normalize each row, 1 to normalize each column
    method : str
        valid methods are:

        -  "vector" : Default for normalization by row (axis=0).
           Normalize along axis as a vector with norm `norm`
        -  "last" : Linear normalization setting last value along the axis to `norm`
        -  "first" : Default for normalization of columns (axis=1).
           Linear normalization setting first value along the given axis to `norm`
        -  "mean" : Normalize so that the mean of each vector along the given axis is `norm`

    norm : float, default 1.0
        Target value of normalization.

    labels : DataFrame
        labels may be passed as keyword argument, in which
        case the label values will also be normalized and returned

    Returns
    -----------
    out : DataFrame or tuple of 2 DataFrames
    
    Normalized data_frame if no labels are provided. Otherwise, a tuple
    containing first normalized data_frame, then normalized labels.

    Notes
    -----------
    If labels are real-valued, they should also be normalized.

    .. Having row_norms as a numpy array should be benchmarked against 
       using a DataFrame:
       http://stackoverflow.com/questions/12525722/normalize-data-in-pandas
       Note: This isn't a bottleneck. Using a feature set with 13k rows and 256
       data_frame ('ge' from 1962 until now), the normalization was immediate.
    """
    norm = kwargs.get('norm', 1.0)
    axis = kwargs.get('axis', 0)
    if axis == 0:
        norm_vector = _get_norms_of_rows(data_frame, kwargs.get('method', 'vector'))
    else:
        norm_vector = _get_norms_of_cols(data_frame, kwargs.get('method', 'first'))

    if 'labels' in kwargs:
        if axis == 0:
            return data_frame.apply(lambda col: col * norm / norm_vector, axis=0), \
                    kwargs['labels'].apply(lambda col: col * norm / norm_vector, axis=0)
        else:
            raise ValueError("label normalization incompatible with normalization by column")
    else:
        if axis == 0:
            return data_frame.apply(lambda col: col * norm / norm_vector, axis=0)
        else:
            return data_frame.apply(lambda row: row * norm / norm_vector, axis=1)

def _get_norms_of_rows(data_frame, method):
    """ return a column vector containing the norm of each row """
    if method == 'vector':
        norm_vector = np.linalg.norm(data_frame.values, axis=1)
    elif method == 'last':
        norm_vector = data_frame.iloc[:, -1].values
    elif method == 'mean':
        norm_vector = np.mean(data_frame.values, axis=1)
    elif method == 'first':
        norm_vector = data_frame.iloc[:, 0].values
    else:
        raise ValueError("no normalization method '{0}'".format(method))
    return norm_vector

def _get_norms_of_cols(data_frame, method):
    """ return a row vector containing the norm of each column """
    if method == 'first':
        norm_vector = data_frame.iloc[0, :].values
    elif method == 'mean':
        norm_vector = np.mean(data_frame.values, axis=0)
    elif method == 'last':
        norm_vector = data_frame.iloc[-1, :].values
    elif method == 'vector':
        norm_vector = np.linalg.norm(data_frame.values, axis=0)
    else:
        raise ValueError("no normalization method '{0}'".format(method))
    return norm_vector



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Options - price (:mod:`pynance.opt.price`)
==================================================

.. currentmodule:: pynance.opt.price
"""

from __future__ import absolute_import

import pandas as pd

from . import constants

def get(optdata, opttype, strike, expiry, showtimeval=True):
    """
    Retrieve price and time value of an option.

    Parameters
    ----------
    optdata : DataFrame
        Data returned from `pn.opt.get()`

    opttype : str {'call', 'put'}
        Type of option on which to collect data.

    strike : numeric
        Strike price.

    expiry : date or date str (e.g. '2015-01-01')
        Expiration date.

    showtimeval : bool, optional
        Whether or not to include time value calculation. Setting
        this value to false will save time if the function is potentially
        being used inside a loop. Defaults to True.

    Returns
    ----------
    opt : float
        Price of option (midpoint between bid and ask).

    eq : float
        Price of underlying.

    qt : datetime.datetime
        Time of quote.

    tv : float, optional
        Time value of option. Returned only as long as `showtimeval` is set
        to True.
    """
    _optrow = _relevant_rows(optdata, (strike, expiry, opttype,),
            "No key for {} strike {} {}".format(expiry, strike, opttype))
    _opt_price = _getprice(_optrow)
    _underlying_price = round(_optrow.loc[:, 'Underlying_Price'].values[0],
            constants.NDIGITS_SIG)
    _qt = pd.to_datetime(_optrow.loc[:, 'Quote_Time'].values[0]).to_datetime()
    if showtimeval:
        if opttype == 'put':
            _timevalue = _get_put_time_val(_opt_price, strike, _underlying_price)
        else:
            _timevalue = _get_call_time_val(_opt_price, strike, _underlying_price)
        return _opt_price, _underlying_price, _qt, _timevalue
    return _opt_price, _underlying_price, _qt

def _relevant_rows(optdata, multikey, errmsg):
    try:
        return optdata.loc[multikey, :]
    except (TypeError, KeyError):
        # message generated by pandas is cryptic
        raise KeyError(errmsg)

def allstrikes(optdata, opttype, expiry):
    """
    Retrieve option prices for all strikes of a given type with a given expiration.

    Parameters
    ----------
    optdata : DataFrame

    opttype : str {'call', 'put'}

    expiry : date or date str

    Returns
    ----------
    df : DataFrame

    eq : float
        Price of underlying.

    qt : datetime.datetime
        Time of quote.
    """
    _relevant = _relevant_rows(optdata, (slice(None), expiry, opttype,),
            "No key for {} {}".format(expiry, opttype))
    _index = _relevant.index.get_level_values('Strike')
    _columns = ['Price', 'Time_Val', 'Last', 'Bid', 'Ask', 'Vol', 'Open_Int']
    _df = pd.DataFrame(index=_index, columns=_columns)
    _underlying = round(_relevant.loc[:, 'Underlying_Price'].values[0],
            constants.NDIGITS_SIG)
    _quotetime = pd.to_datetime(_relevant.loc[:, 'Quote_Time'].values[0]).to_datetime()
    for _col in _columns[2:]:
        _df.loc[:, _col] = _relevant.loc[:, _col].values
    _df.loc[:, 'Price'] = (_df.loc[:, 'Bid'] + _df.loc[:, 'Ask']) / 2.
    _set_tv_strike_ix(_df, opttype, 'Price', 'Time_Val', _underlying)
    return _df, _underlying, _quotetime

def allexpiries(optdata, opttype, strike):
    """
    Prices for given strike on all available dates.

    Parameters
    ----------
    optdata : DataFrame

    opttype : str {'call', 'put'}

    strike : numeric

    Returns
    ----------
    df : DataFrame

    eq : float
        Price of underlying.

    qt : datetime.datetime
        Time of quote.
    """
    _relevant = _relevant_rows(optdata, (strike, slice(None), opttype,),
            "No key for {} {}".format(strike, opttype))
    _index = _relevant.index.get_level_values('Expiry')
    _columns = ['Price', 'Time_Val', 'Last', 'Bid', 'Ask', 'Vol', 'Open_Int']
    _df = pd.DataFrame(index=_index, columns=_columns)
    _underlying = round(_relevant.loc[:, 'Underlying_Price'].values[0],
            constants.NDIGITS_SIG)
    _quotetime = pd.to_datetime(_relevant.loc[:, 'Quote_Time'].values[0]).to_datetime()
    for _col in _columns[2:]:
        _df.loc[:, _col] = _relevant.loc[:, _col].values
    _df.loc[:, 'Price'] = (_df.loc[:, 'Bid'] + _df.loc[:, 'Ask']) / 2.
    _set_tv_other_ix(_df, opttype, 'Price', 'Time_Val', _underlying, strike)
    return _df, _underlying, _quotetime

def _set_tv_other_ix(df, opttype, pricecol, tvcol, eqprice, strike):
    if opttype == 'put':
        if strike <= eqprice:
            df.loc[:, tvcol] = df.loc[:, pricecol]
        else:
            _diff = eqprice - strike
            df.loc[:, tvcol] = df.loc[:, pricecol] + _diff
    else:
        if eqprice <= strike:
            df.loc[:, tvcol] = df.loc[:, pricecol]
        else:
            _diff = strike - eqprice
            df.loc[:, tvcol] = df.loc[:, pricecol] + _diff

def _set_tv_strike_ix(df, opttype, pricecol, tvcol, eqprice):
    df.loc[:, tvcol] = df.loc[:, pricecol]
    if opttype == 'put':
        _mask = (df.index > eqprice)
        df.loc[_mask, tvcol] += eqprice - df.index[_mask]
    else:
        _mask = (df.index < eqprice)
        df.loc[_mask, tvcol] += df.index[_mask] - eqprice
    return

def _getprice(optrow):
    _bid = optrow.loc[:, 'Bid'].values[0]
    _ask = optrow.loc[:, 'Ask'].values[0]
    return round((_bid + _ask) / 2., constants.NDIGITS_SIG)

def _get_put_time_val(putprice, strike, eqprice):
    if strike <= eqprice:
        return putprice
    return round(putprice + eqprice - strike, constants.NDIGITS_SIG)
    
def _get_call_time_val(callprice, strike, eqprice):
    if eqprice <= strike:
        return callprice
    return round(callprice + strike - eqprice, constants.NDIGITS_SIG)



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Common - generic functions (:mod:`pynance.common`)
==================================================

.. currentmodule:: pynance.common
"""

import pandas as pd

def featurize(equity_data, n_sessions, **kwargs):
    """
    Generate a raw (unnormalized) feature set from the input data.
    The value at `column` on the given date is taken
    as a feature, and each row contains values for n_sessions

    Parameters
    -----------
    equity_data : DataFrame
        data from which to generate features

    n_sessions : int
        number of sessions to use as features

    selection : str, default: 'Adj Close'
        column of `equity_data` from which to generate features.

    columns : list, default: ``map(str, range((-n_sessions + 1), 1))``
        column names for output DataFrame. Default will look like:
        ['-5', '-4', '-3', '-2', '-1', '0'].

    Returns
    ----------
    out : DataFrame
        Each row is a sequence of `n_sessions` session values where
        the last column matches the value on the date specified by
        the DataFrame index.

    Examples
    --------
    >>> pn.featurize(equity_data, n_sessions, **kwargs)
    """
    #Benchmarking
    #>>> s = 'from __main__ import data\nimport datetime as dt\n'
    #>>> timeit.timeit('data.featurize(data.get("ge", dt.date(1960, 1, 1), 
    #        dt.date(2014, 12, 31)), 256)', setup=s, number=1)
    #1.6771750450134277
    columns = kwargs.get('columns', map(str, range(-n_sessions + 1, 1)))
    selection = kwargs.get('selection', 'Adj Close')
    # empty DataFrame with desired index and column labels
    features = pd.DataFrame(index=equity_data.index[(n_sessions - 1):],
            columns=columns, dtype='float64')
    values = equity_data[selection].values
    for i in range(n_sessions - 1):
        features.iloc[:, i] = values[i:(-n_sessions + i + 1)]
    features.iloc[:, n_sessions - 1] = values[(n_sessions - 1):]
    return features

def decorate(fn, *args, **kwargs):
    """
    Return a new function that replicates the behavior of the input
    but also returns an additional value. Used for creating functions
    of the proper type to pass to `labeledfeatures()`.

    Parameters
    ----------
    fn : function

    *args : any
        Additional parameters that the returned function will return

    **kwargs : dict
        Each element in `kwargs` will become an attribute of the output
        function.

    Returns
    ----------
    wrapped : function
        New function that acts like `fn` except that it also returns
        an additional value.

    Examples
    ----------
    >>> from functools import partial
    >>> forecast_interval = 32
    >>> features, labels = pn.data.labeledfeatures(eqdata, 256, featurefn,
    ...        decorate(partial(pn.data.lab.growth, forecast_interval, 'Adj Close'), forecast_interval))
    >>> def f():
    ...    return 0, 1 
    ...
    >>> pn.decorate(f, 3, 4, 5)()
    (0, 1, 3, 4, 5)
    >>> pn.decorate(lambda x: x * .5, 3, 4, 5)(1.)
    (1., 3, 4, 5)
    >>> pn.decorate(lambda x: x, 1 2)('foo')
    ('foo', 1, 2)
    >>> pn.decorate(f, 'foo'):
    (0, 1, 'foo')
    pn.decorate(f, 0, foo='bar').foo
    >>> 'bar'

    Notes
    ----------
    If `fn` returns multiple values, these will be returned in sequence
    as the first values returned by `add_rets(fn, arg0, arg1, arg2)`. See example
    above.
    """
    def _wrapper(*_args, **kwargs):
        _ret = fn(*_args, **kwargs)
        if isinstance(_ret, tuple):
            return _ret + args
        if len(args) == 0:
            return _ret
        return (_ret,) + args
    for key, value in kwargs.items():
        _wrapper.__dict__[key] = value
    return _wrapper

def expand(fn, col, inputtype=pd.DataFrame):
    """
    Wrap a function applying to a single column to make a function
    applying to a multi-dimensional dataframe or ndarray

    Parameters
    ----------
    fn : function
        Function that applies to a series or vector.

    col : str or int
        Index of column to which to apply `fn`.

    inputtype : class or type
        Type of input to be expected by the wrapped function.
        Normally pd.DataFrame or np.ndarray. Defaults to pd.DataFrame.

    Returns
    ----------
    wrapped : function
        Function that takes an input of type `inputtype` and applies
        `fn` to the specified `col`.
    """
    if inputtype == pd.DataFrame:
        if isinstance(col, int):
            def _wrapper(*args, **kwargs):
                return fn(args[0].iloc[:, col], *args[1:], **kwargs)
            return _wrapper
        def _wrapper(*args, **kwargs):
            return fn(args[0].loc[:, col], *args[1:], **kwargs)
        return _wrapper
    elif inputtype == np.ndarray:
        def _wrapper(*args, **kwargs):
            return fn(args[0][:, col], *args[1:], **kwargs)
        return _wrapper
    raise TypeError("invalid input type")

def has_na(eqdata):
    """
    Return false if `eqdata` contains no missing values.

    Parameters
    ----------
    eqdata : DataFrame or ndarray
        Data to check for missing values (NaN, None)

    Returns
    ----------
    answer : bool
        False iff `eqdata` contains no missing values.
    """
    if isinstance(eqdata, pd.DataFrame):
        _values = eqdata.values
    else:
        _values = eqdata
    return len(_values[pd.isnull(_values)]) > 0



"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Data - remote retrieval (:mod:`pynance.data.retrieve`)
=========================================================

.. currentmodule:: pynance.data.retrieve

Wraps `Pandas Remote Data Access 
<http://pandas.pydata.org/pandas-docs/stable/remote_data.html>`_.
"""

import pandas.io.data as web

def get(equity, start, end):
    """ 
    Get DataFrame for an individual equity from Yahoo!  
    
    Examples
    --------
    >>> import pynance as pn
    >>> aapl = pn.data.get('aapl', '2014-03-01', '2015-03-01')
    >>> goog = pn.data.get('goog', '2014', '2015')
    """
    return web.DataReader(equity, 'yahoo', start, end)



"""
.. Copyright (c) 2014 Marshall Farrier
   license http://opensource.org/licenses/MIT

Data (:mod:`pynance.data`)
===============================================

.. currentmodule:: pynance.data

:mod:`pynance.data.combine`

:mod:`pynance.data.feat`

:mod:`pynance.data.lab`

:mod:`pynance.data.prep`

:mod:`pynance.data.retrieve`
"""

from __future__ import absolute_import

__all__ = ["combine", "feat", "lab", "prep", "retrieve"]

# imported directly into data module
from . import combine
from .combine import *
from . import prep
from .prep import *
from . import retrieve
from .retrieve import *

# imported as submodule
from . import feat
from . import lab



"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Options (:mod:`pynance.opt`)
===============================================

.. currentmodule:: pynance.opt

:mod:`pynance.opt.covcall`

:mod:`pynance.opt.price`

:mod:`pynance.opt.retrieve`

:mod:`pynance.opt.spread`
"""
from __future__ import absolute_import

__all__ = ["covcall", "price", "retrieve", "spread"]

# imported directly into module
from . import retrieve
from .retrieve import *

# imported as submodule
from . import covcall
from . import price
from . import spread



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Interest rates (:mod:`pynance.interest`)
========================================

.. currentmodule:: pynance.interest
"""

import datetime as dt
import math

import pandas as pd

def yrlygrowth(total_growth, years):
    """
    Determine the annual growth from the growth over an
    arbitrary time span.
    """
    return math.exp(math.log(total_growth) / years)

def yrlyret(total_interest, years):
    """
    Determine annual interest from interest compounded
    over an arbitrary time span.
    """
    return yrlygrowth(total_interest + 1.0, years) - 1.0

def growthfromrange(rangegrowth, startdate, enddate):
    """
    Annual growth given growth from start date to end date.
    """
    _yrs = (pd.Timestamp(enddate) - pd.Timestamp(startdate)).total_seconds() /\
            dt.timedelta(365.25).total_seconds()
    return yrlygrowth(rangegrowth, _yrs)

def retfromrange(rangeret, startdate, enddate):
    """
    Annual return given return from start date to end date.
    """
    return growthfromrange(1. + rangeret, startdate, enddate) - 1.

def growthtocont(annualgr):
    """
    Convert annual growth to continuous compounding rate
    """
    return math.log(annualgr)

def conttogrowth(contrate):
    """
    Convert continuous compounding rate to annual growth
    """
    return math.exp(contrate)

def compgrowth(annual_growth, years):
    """
    Compound `annual_growth` over given `years`.
    """
    return annual_growth**years

def compret(annual_interest, years):
    """
    Compound `annual_interest` over given `years`
    """
    return (1.0 + annual_interest)**years - 1.0

def pvannuity(ret, n):
    """
    Present value of n constant payments assuming
    an interest rate per payment interval of `ret`
    """
    return (1. - (1. + ret)**-n) / ret

def loanpayment(amount, rate, n_payments):
    """
    Amount of a constant loan payment.

    Parameters
    ----------
    amount : float
        amount of loan
    rate : float
        interest rate for the given payment, i.e., monthly
        interest if the loan is to be paid off monthly.
    n_payments : int
        number of payments to be made.

    Returns
    -------
    out : float
        amount to be paid each period.
    """
    return float(amount) / pvannuity(rate, n_payments)



"""
.. Copyright (c) 2014 Marshall Farrier
   license http://opensource.org/licenses/MIT

Data - combine features and labels (:mod:`pynance.data.combine`)
==================================================================

.. currentmodule:: pynance.data.combine
"""

import pandas as pd

from . import feat

def labeledfeatures(eqdata, featurefunc, labelfunc):
    """
    Return features and labels for the given equity data.

    Each row of the features returned contains `2 * n_sessions + 1` columns
    (or 1 less if the constant feature is excluded). After the constant feature,
    if present, there will be `n_sessions` columns derived from daily growth
    of the given price column, which defaults to 'Adj Close'. There will then
    follow another `n_sessions` columns representing current volume as
    a multiple of average volume over the previous 252 (or other value determined
    by the user) sessions. 
    
    The returned features are not centered or normalized because these
    adjustments need to be made after test or cross-validation data has
    been removed.

    The constant feature is prepended by default.

    The labels are derived from `eqdata` using `labelfunc`.

    Parameters
    ----------
    eqdata : DataFrame
        Expected is a dataframe as return by the `get()` function. A column
        labeled 'Volume' must be present.

    featurefunc : function
        Function taking a dataframe of simple equity data as argument
        and returning a dataframe of features and an integer representing
        the number of rows that had to be skipped at the beginning of
        the index of the input dataframe. The rows skipped are used
        to synchronize the indices of features and labels. For example,
        if the features are composed of 4 successive daily returns, then
        the date of row 0 of features would be the same as the date of row 3
        (counting from 0) of input data. So the corresponding `featurefunc`
        would return a dataframe and the value 3.

    labelfunc : function
        function for deriving labels from `eqdata`. `labelfunc` must
        take a single argument: `df`, a dataframe to which `labelfunc` will be applied.
        `labelfunc` should return a dataframe of labels followed by an int
        specifying the number of feature rows to skip at the end of the feature
        dataframe. For example, if features are relative prices 64 days out,
        these features will only be known up until 64 days before the data
        runs out. In order to properly align features and labels, the features
        should not include the last 64 rows that would otherwise be possible.

        Usage:
        `labels, skipatend = labelfunc(eqdata)`

    Returns
    -------
    features : DataFrame
        The features derived from the given parameters.

    labels : DataFrame
        The labels derived from the given parameters.
    """
    _size = len(eqdata.index)
    _labels, _skipatend = labelfunc(eqdata)
    _features, _skipatstart = featurefunc(eqdata.iloc[:(_size - _skipatend), :])
    return _features, _labels.iloc[_skipatstart:, :]



"""
.. Copyright (c) 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Options - remote retrieval (:mod:`pynance.opt.retrieve`)
=========================================================

.. currentmodule:: pynance.opt.retrieve
"""

from __future__ import absolute_import

import pandas
from pandas.io.data import Options

from . import constants

def get(equity, showinfo=True):
    """
    Retrieve all current options chains for given equity.

    Parameters
    -------------
    equity : str
        Equity for which to retrieve options data.

    showinfo : bool, optional
        If true (default), available expiries and equity price are 
        printed to console. 

    Returns
    -------------
    optdata : DataFrame
        All options data for given equity currently available
        from Yahoo! Finance.

    expdates : pandas.tseries.index.DatetimeIndex
        Index of all active expiration dates.

    eqprice : float
        Price of underlying equity at the time options data
        was retrieved.
        
    Notes
    -------------
    For convenience, expiration dates are shown by default. So you
    don't have to call other functions to figure out what the exact expiration
    dates are for the various options you might be considering.
    An index of the expiration timestamps is also returned,
    so that you can then pass dates not only in the form
    '2015-08-21' (or other object are formats that pandas
    can convert to timestamps) but also as a numerical index
    in pandas timeseries (cf. example below).

    To disable this feature and return a single dataframe of options
    data, set the `showinfo` argument to False.

    Examples
    -------------
    >>> fopt, fexp, feq = pn.opt.get('f')
    Expirations:
    ...
    >>> fstraddle = pn.opt.spread.straddle(fopt, 16, fexp[4])
    """
    _optmeta = Options(equity, 'yahoo')
    _optdata = None
    try:
        _optdata = _optmeta.get_all_data()
    except (AttributeError, ValueError, pandas.io.data.RemoteDataError):
        raise pandas.io.data.RemoteDataError(
                "No options data available for {!r}".format(equity))
    if showinfo:
        print("Expirations:")
        showexpiries(_optdata)
        print("Stock: {:.2f}".format(_optdata.iloc[0].loc['Underlying_Price']))
    return _optdata, _optdata.index.levels[1], round(_optdata.iloc[0].loc['Underlying_Price'],
            constants.NDIGITS_SIG)

def getexpiries(optdata):
    """
    Get all expiration dates contained in the data index.

    Parameters
    -------------
    optdata : DataFrame
        Collection of options data as retrieved from `pn.opt.get()`

    Returns
    -------------
    expdates : pandas.tseries.index.DatetimeIndex
        Index of all active expiration dates.
    """
    return optdata.index.levels[1]

def showexpiries(optdata):
    """
    Show all expiration dates but return nothing

    Parameters
    -------------
    optdata : DataFrame
        Collection of options data as retrieved from `pn.opt.get()`
    """
    _i = 0
    for _datetime in optdata.index.levels[1].to_pydatetime():
        print("{:2d} {}".format(_i, _datetime.strftime('%Y-%m-%d')))
        _i += 1



"""
.. Copyright (c) 2014 Marshall Farrier
   license http://opensource.org/licenses/MIT

Technical analysis (:mod:`pynance.tech`)
===============================================

.. currentmodule:: pynance.tech

:mod:`pynance.tech.movave`

:mod:`pynance.tech.simple`
"""

from __future__ import absolute_import

_all__ = ["movave", "simple"]

# import directly into tech module
from . import movave
from .movave import *
from . import simple
from .simple import *



from __future__ import division, absolute_import, print_function

__all__ = ["common", "data", "chart", "interest", "opt", "pf", "tech"]

from . import common
from .common import *
from . import data
from . import chart
from . import interest
from . import opt
from . import pf
from . import tech



"""
.. Copyright (c) 2014- Marshall Farrier
   license http://opensource.org/licenses/MIT

Technical analysis - moving averages (:mod:`pynance.tech.movave`)
==================================================================

.. currentmodule:: pynance.tech.movave
"""

from __future__ import absolute_import

import numpy as np
import pandas as pd

from . import simple

def sma(eqdata, **kwargs):
    """ 
    simple moving average 

    Parameters
    ----------
    eqdata : DataFrame
    window : int, optional
        Lookback period for sma. Defaults to 20.
    outputcol : str, optional
        Column to use for output. Defaults to 'SMA'.
    selection : str, optional
        Column of eqdata on which to calculate sma. If
        `eqdata` has only 1 column, `selection` is ignored,
        and sma is calculated on that column. Defaults
        to 'Adj Close'.
    """
    if len(eqdata.shape) > 1 and eqdata.shape[1] != 1:
        _selection = kwargs.get('selection', 'Adj Close')
        _eqdata = eqdata.loc[:, _selection]
    else:
        _eqdata = eqdata
    _window = kwargs.get('window', 20)
    _outputcol = kwargs.get('outputcol', 'SMA')
    ret = pd.DataFrame(index=_eqdata.index, columns=[_outputcol], dtype=np.float64)
    ret.loc[:, _outputcol] = pd.rolling_mean(_eqdata, window=_window).values.flatten()
    return ret

def ema(eqdata, **kwargs):
    """
    Exponential moving average with the given span.

    Parameters
    ----------
    eqdata : DataFrame
        Must have exactly 1 column on which to calculate EMA
    span : int, optional
        Span for exponential moving average. Cf. `pandas.stats.moments.ewma 
        <http://pandas.pydata.org/pandas-docs/stable/generated/pandas.stats.moments.ewma.html>`_ and
        `additional Pandas documentation 
        <http://pandas.pydata.org/pandas-docs/stable/computation.html#exponentially-weighted-moment-functions>`_.
    outputcol : str, optional
        Column to use for output. Defaults to 'EMA'.
    selection : str, optional
        Column of eqdata on which to calculate ema. If
        `eqdata` has only 1 column, `selection` is ignored,
        and ema is calculated on that column. Defaults
        to 'Adj Close'.

    Returns
    ---------
    emadf : DataFrame
        Exponential moving average using the given `span`.
    """
    if len(eqdata.shape) > 1 and eqdata.shape[1] != 1:
        _selection = kwargs.get('selection', 'Adj Close')
        _eqdata = eqdata.loc[:, _selection]
    else:
        _eqdata = eqdata
    _span = kwargs.get('span', 20)
    _col = kwargs.get('outputcol', 'EMA')
    # necessary because pd.ewma() outputs Series rather than DataFrame
    _emadf = pd.DataFrame(index=_eqdata.index, columns=[_col], dtype=np.float64)
    _emadf.loc[:, _col] = pd.ewma(_eqdata, span=_span).values.flatten()
    return _emadf

def ema_growth(eqdata, **kwargs):
    """
    Growth of exponential moving average.

    Parameters
    ----------
    eqdata : DataFrame
    span : int, optional
        Span for exponential moving average. Defaults to 20.
    outputcol : str, optional.
        Column to use for output. Defaults to 'EMA Growth'.
    selection : str, optional
        Column of eqdata on which to calculate ema growth. If
        `eqdata` has only 1 column, `selection` is ignored,
        and ema growth is calculated on that column. Defaults
        to 'Adj Close'.

    Returns
    ---------
    out : DataFrame
        Growth of exponential moving average from one day to next
    """
    _growth_outputcol = kwargs.get('outputcol', 'EMA Growth')
    _ema_outputcol = 'EMA'
    kwargs['outputcol'] = _ema_outputcol
    _emadf = ema(eqdata, **kwargs)
    return simple.growth(_emadf, selection=_ema_outputcol, outputcol=_growth_outputcol)

def volatility(eqdata, **kwargs):
    """
    Volatility (standard deviation) over the given window

    Parameters
    ----------
    eqdata : DataFrame
    window : int, optional
        Lookback period. Defaults to 20.
    outputcol : str, optional
        Name of column to be used in returned dataframe. Defaults to 'Risk'.
    selection : str, optional
        Column of eqdata on which to calculate volatility. If
        `eqdata` has only 1 column, `selection` is ignored,
        and volatility is calculated on that column. Defaults
        to 'Adj Close'.

    Returns
    ---------
    risk : DataFrame
        Moving volatility with the given lookback.
    """
    if len(eqdata.shape) > 1 and eqdata.shape[1] != 1:
        _selection = kwargs.get('selection', 'Adj Close')
        _eqdata = eqdata.loc[:, _selection]
    else:
        _eqdata = eqdata
    _window = kwargs.get('window', 20)
    _colname = kwargs.get('outputcol', 'Risk')
    _risk = pd.DataFrame(index=_eqdata.index, columns=[_colname], dtype=np.float64)
    _risk.loc[:, _colname] = pd.rolling_std(_eqdata, window=_window).values.flatten()
    return _risk

def growth_volatility(eqdata, **kwargs):
    """
    Return the volatility of growth.

    Note that, like :func:`pynance.tech.simple.growth` but in contrast to 
    :func:`volatility`, :func:`growth_volatility`
    applies directly to a dataframe like that returned by 
    :func:`pynance.data.retrieve.get`, not necessarily to a single-column dataframe.

    Parameters
    ----------
    eqdata : DataFrame
        Data from which to extract growth volatility. An exception
        will be raised if `eqdata` does not contain a column 'Adj Close'
        or an optional name specified by the `selection` parameter.
    window : int, optional
        Window on which to calculate volatility. Defaults to 20.
    selection : str, optional
        Column of eqdata on which to calculate volatility of growth. Defaults
        to 'Adj Close'
    outputcol : str, optional
        Column to use for output. Defaults to 'Growth Risk'.

    Returns
    ---------
    out : DataFrame
        Dataframe showing the volatility of growth over the specified `window`.
    """
    _window = kwargs.get('window', 20)
    _selection = kwargs.get('selection', 'Adj Close')
    _outputcol = kwargs.get('outputcol', 'Growth Risk')
    _growthdata = simple.growth(eqdata, selection=_selection)
    return volatility(_growthdata, outputcol=_outputcol, window=_window)

def bollinger(eqdata, **kwargs):
    """ 
    Bollinger bands

    Returns bolldf, smadf where bolldf is a DataFrame containing
    Bollinger bands with columns 'Upper' and 'Lower' and smadf contains
    the simple moving average.

    Parameters
    ----------
    eqdata : DataFrame
        Must include a column specified in the `selection` parameter or, 
        if no `selection` parameter is given, a column 'Adj Close'.
    window : int, optional
        Lookback period.
    multiple : float, optional
        Multiple of standard deviation above and below sma to use
        in calculating band value. Defaults to 2.0.
    selection : str, optional
        Column of `eqdata` on which to calculate bollinger bands.
        Defaults to 'Adj Close'.

    Returns
    ---------
    bolldf : DataFrame
        Dataframe containing columns 'Upper' and 'Lower' describing
        the given multiple of standard deviations above and below
        simple moving average for the lookback period.
    smadf : DataFrame
        Simple moving average given the specified lookback.
    """
    _window = kwargs.get('window', 20)
    _multiple = kwargs.get('multiple', 2.)
    _selection = kwargs.get('selection', 'Adj Close')
    # ensures correct name for output column of sma()
    kwargs['outputcol'] = 'SMA'
    _smadf = sma(eqdata, **kwargs)
    _sigmas = pd.rolling_std(eqdata.loc[:, _selection], window=_window).values.flatten()
    _diff = _multiple * _sigmas
    _bolldf = pd.DataFrame(index=eqdata.index, columns=['Upper', 'Lower'], dtype=np.float64)
    _bolldf.loc[:, 'Upper'] = _smadf.iloc[:, 0].values + _diff
    _bolldf.loc[:, 'Lower'] = _smadf.iloc[:, 0].values - _diff
    return _bolldf, _smadf

def ratio_to_ave(window, eqdata, **kwargs):
    """
    Return values expressed as ratios to the average over some number
    of prior sessions.

    Parameters
    ----------
    eqdata : DataFrame
        Must contain a column with name matching `selection`, or, if
        `selection` is not specified, a column named 'Volume'
    window : int
        Interval over which to calculate the average. Normally 252 (1 year)
    selection : str, optional
        Column to select for calculating ratio. Defaults to 'Volume'
    skipstartrows : int, optional
        Rows to skip at beginning in addition to the `window` rows
        that must be skipped to get the baseline volume. Defaults to 0.
    skipendrows : int, optional
        Rows to skip at end. Defaults to 0.
    outputcol : str, optional
        Name of column in output dataframe. Defaults to 'Ratio to Ave'

    Returns
    ---------
    out : DataFrame
    """
    _selection = kwargs.get('selection', 'Volume')
    _skipstartrows = kwargs.get('skipstartrows', 0)
    _skipendrows = kwargs.get('skipendrows', 0)
    _outputcol = kwargs.get('outputcol', 'Ratio to Ave')
    _size = len(eqdata.index)
    _eqdata = eqdata.loc[:, _selection]

    _sma = pd.rolling_mean(_eqdata.iloc[:-1 - _skipendrows], window=window).values
    _outdata = _eqdata.values[window + _skipstartrows:_size - _skipendrows] /\
            _sma[window + _skipstartrows - 1:]
    _index = eqdata.index[window + _skipstartrows:_size - _skipendrows]
    return pd.DataFrame(_outdata, index=_index, columns=[_outputcol], dtype=np.float64)



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Portfolio optimization (:mod:`pynance.pf`)
==================================================

.. currentmodule:: pynance.pf
"""

import numpy as np

def optimize(exp_rets, covs):
    """
    Return parameters for portfolio optimization.

    Parameters
    ----------
    exp_rets : ndarray
        Vector of expected returns for each investment..
    covs : ndarray
        Covariance matrix for the given investments.

    Returns
    ---------
    a : ndarray
        The first vector (to be combined with target return as scalar)
        in the linear equation for optimal weights.
    b : ndarray
        The second (constant) vector in the linear equation for
        optimal weights.
    least_risk_ret : int
        The return achieved on the portfolio that combines the given
        equities so as to achieve the lowest possible risk.

    Notes
    ---------
    *   The length of `exp_rets` must match the number of rows
        and columns in the `covs` matrix.
    *   The weights for an optimal portfolio with expected return
        `ret` is given by the formula `w = ret * a + b` where `a`
        and `b` are the vectors returned here. The weights `w` for
        the portfolio with lowest risk are given by `w = least_risk_ret * a + b`.
    *   An exception will be raised if the covariance matrix
        is singular or if each prospective investment has the
        same expected return.
    """
    _cov_inv = np.linalg.inv(covs)        

    # unit vector
    _u = np.ones((len(exp_rets)))

    # compute some dot products one time only
    _u_cov_inv = _u.dot(_cov_inv)
    _rets_cov_inv = exp_rets.dot(_cov_inv)

    # helper matrix for deriving Lagrange multipliers
    _m = np.empty((2, 2))
    _m[0, 0] = _rets_cov_inv.dot(exp_rets)
    _m[0, 1] = _u_cov_inv.dot(exp_rets)
    _m[1, 0] = _rets_cov_inv.dot(_u)
    _m[1, 1] = _u_cov_inv.dot(_u)

    # compute values to return
    _m_inv = np.linalg.inv(_m)
    a = _m_inv[0, 0] * _rets_cov_inv + _m_inv[1, 0] * _u_cov_inv
    b = _m_inv[0, 1] * _rets_cov_inv + _m_inv[1, 1] * _u_cov_inv
    least_risk_ret = _m[0, 1] / _m[1, 1]
    return a, b, least_risk_ret



"""
For instructions on how to put a package on PyPI:
http://peterdowns.com/posts/first-time-with-pypi.html

To upload a new version to PyPI:
1. Udate `VERSION` below. 
2. Commit all changes. 
3. Create a tag for new version in git:
    % git tag 0.0.1 -m "Fixed some problems"
    % git tag # to verify that new tag is in list
    % git push --tags

4. Register and upload to PyPI Test:
% python setup.py register -r pypitest
% python setup.py sdist upload -r pypitest

5. Register and upload to PyPI Live:
% python setup.py register -r pypi
% python setup.py sdist upload -r pypi
"""

from setuptools import setup, find_packages

MAJOR   = 0
MINOR   = 2
MICRO   = 0
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

CLASSIFIERS = [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Topic :: Office/Business",
        "Topic :: Office/Business :: Financial",
        "Topic :: Office/Business :: Financial :: Investment",
        ]

KEYWORDS = [
        "finance",
        "investment",
        "stocks",
        "equities",
        "market",
        "options",
        "derivatives",
        ]

DEPENDENCIES = [
        "numpy",
        "pandas",
        "matplotlib",
        "lxml",
        "html5lib",
        "beautifulsoup4",
        ]

setup(
        name='pynance',
        packages=find_packages(),
        version=VERSION,
        description='Retrieve and analyse financial market data',
        author='Marshall Farrier',
        author_email='marshalldfarrier@gmail.com',
        url='http://pynance.net',
        download_url=('https://github.com/aisthesis/pynance/tarball/' + VERSION),
        keywords=' '.join(KEYWORDS),
        classifiers=CLASSIFIERS,
        install_requires=DEPENDENCIES
        )



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Options - covered calls (:mod:`pynance.opt.covcall`)
======================================================

.. currentmodule:: pynance.opt.covcall
"""

from __future__ import absolute_import

import pandas as pd

from . import constants

def get(stockprice, callprice, strike, shares=1, buycomm=0., excomm=0., dividend=0.):
    """
    Metrics for covered calls.

    Cf. McMillan, Options as a Strategic Investment, 5th ed., p. 43

    Parameters
    ----------
    stockprice : float
        Price at which stock is purchased.

    callprice : float
        Price for which call is sold.

    strike : float
        Strike price of call sold.

    shares : int, optional
        Number of shares of stock. Defaults to 1.

    buycomm : float, optional
        Commission paid on total initial purchase.

    excomm : float optional
        Commission to be paid if option is exercised.

    dividend : float, optional
        Total dividends per share expected between purchase and expiration.

    Returns
    ----------
    metrics : pd.DataFrame
        Investment metrics
    """
    _index = ['Stock Cost', 'Option Premium', 'Commission', 'Total Invested', 'Dividends', 'Stock if Ex', 
            'Comm if Ex', 'Profit if Ex', 'Ret if Ex', 'Profit if Unch', 'Ret if Unch', 'Break_Even Price',
            'Protection Pts', 'Protection Pct']
    _metrics = pd.DataFrame(index=_index, columns=['Value'])
    _shares = float(shares)
    _dividends = _shares * dividend
    _metrics.loc['Stock Cost', 'Value'] = _stockcost = _shares * stockprice
    _metrics.loc['Option Premium', 'Value'] = _optprem = _shares * callprice
    _metrics.loc['Commission', 'Value'] = float(buycomm)
    _metrics.loc['Total Invested', 'Value'] = _invested = _stockcost - _optprem + buycomm
    _metrics.loc['Dividends', 'Value'] = _dividends
    _metrics.loc['Stock if Ex', 'Value'] = _stocksale = strike * _shares
    _metrics.loc['Comm if Ex', 'Value'] = float(excomm)
    _metrics.loc['Profit if Ex', 'Value'] = _profitex = _stocksale + _dividends - _invested - excomm
    _metrics.loc['Ret if Ex', 'Value'] = round(_profitex / _invested, constants.NDIGITS_SIG)
    _metrics.loc['Profit if Unch', 'Value'] = _profitunch = _stockcost + _dividends - _invested
    _metrics.loc['Ret if Unch', 'Value'] = round(_profitunch / _invested, constants.NDIGITS_SIG)
    _metrics.loc['Break_Even Price', 'Value'] = _breakeven = round((_invested - _dividends) / _shares, 
            constants.NDIGITS_SIG)
    _metrics.loc['Protection Pts', 'Value'] = _protpts = stockprice - _breakeven
    _metrics.loc['Protection Pct', 'Value'] = round(_protpts / stockprice, constants.NDIGITS_SIG)
    return _metrics



"""
.. Copyright (c) 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Options - spreads (:mod:`pynance.opt.spread`)
==================================================

.. currentmodule:: pynance.opt.spread
"""

from __future__ import absolute_import

import numpy as np
import pandas as pd

from . import price

def calendar(optdata, opttype, strike, expiry1, expiry2):
    """
    Metrics for evaluating a simple calendar spread.

    Parameters
    ------------
    optdata : DataFrame
        Data returned from :func:`pynance.opt.retrieve.get`
    opttype : str ('call' or 'put')
        Type of option on which to collect data.
    strike : numeric
        Strike price.
    expiry1 : date or date str (e.g. '2015-01-01')
        Earlier expiration date.
    expiry2 : date or date str (e.g. '2015-01-01')
        Later expiration date.

    Returns
    ------------
    metrics : DataFrame
        Metrics for evaluating spread.
    """
    _price1, _underlying, _qt = price.get(optdata, opttype, strike, expiry1, False)
    _price2 = price.get(optdata, opttype, strike, expiry2, False)[0]
    _index = ['Near', 'Far', 'Debit', 'Underlying', 'Quote_Time']
    _vals = np.array([_price1, _price2, _price2 - _price1, _underlying, _qt])
    return pd.DataFrame(_vals, index=_index, columns=['Value'])

def dblcal(optdata, lowstrike, highstrike, expiry1, expiry2):
    """
    Metrics for evaluating a double calendar spread.

    Parameters
    ------------
    optdata : DataFrame
        Data returned from :func:`pynance.opt.retrieve.get`
    opttype : str ('call' or 'put')
        Type of option on which to collect data.
    lowstrike : numeric
        Lower strike price. To be used for put spread.
    highstrike : numeric
        Higher strike price. To be used for call spread.
    expiry1 : date or date str (e.g. '2015-01-01')
        Earlier expiration date.
    expiry2 : date or date str (e.g. '2015-01-01')
        Later expiration date.

    Returns
    ------------
    metrics : DataFrame
        Metrics for evaluating spread.
    """
    _index = ['Near Call', 'Far Call', 'Call Ratio', 'Near Put', 'Far Put', 
            'Put Ratio', 'Near to Far Ratio', 'Debit', 'Underlying', 'Quote_Time']
    _metrics = pd.DataFrame(index=_index, columns=['Value'])
    _nearcall, _metrics.loc['Underlying', 'Value'], _metrics.loc['Quote_Time', 'Value'] =\
            price.get(optdata, 'call', highstrike, expiry1, False)
    _metrics.loc['Near Call', 'Value'] = _nearcall
    _farcall = price.get(optdata, 'call', highstrike, expiry2, False)[0]
    _metrics.loc['Far Call', 'Value'] = _farcall
    _metrics.loc['Call Ratio', 'Value'] = _nearcall / _farcall
    _nearput = price.get(optdata, 'put', lowstrike, expiry1, False)[0]
    _metrics.loc['Near Put', 'Value'] = _nearput
    _farput = price.get(optdata, 'put', lowstrike, expiry2, False)[0]
    _metrics.loc['Far Put', 'Value'] = _farput
    _metrics.loc['Put Ratio', 'Value'] = _nearput / _farput
    _metrics.loc['Near to Far Ratio', 'Value'] = (_nearcall + _nearput) / (_farcall + _farput)
    _metrics.loc['Debit', 'Value'] = _farcall + _farput - _nearcall - _nearput
    return _metrics

def diagbtrfly(optdata, lowstrike, midstrike, highstrike, expiry1, expiry2):
    """
    Metrics for evaluating a diagonal butterfly spread.

    Parameters
    ------------
    optdata : DataFrame
        Data returned from :func:`pynance.opt.retrieve.get`
    opttype : str ('call' or 'put')
        Type of option on which to collect data.
    lowstrike : numeric
        Lower strike price. To be used for far put.
    midstrike : numeric
        Middle strike price. To be used for near straddle.
        Typically at the money.
    highstrike : numeric
        Higher strike price. To be used for far call.
    expiry1 : date or date str (e.g. '2015-01-01')
        Earlier expiration date.
    expiry2 : date or date str (e.g. '2015-01-01')
        Later expiration date.

    Returns
    ------------
    metrics : DataFrame
        Metrics for evaluating spread.
    """
    _index = ['Straddle Call', 'Straddle Put', 'Straddle Total', 'Far Call', 'Far Put', 'Far Total',
            'Straddle to Far Ratio', 'Credit', 'Underlying', 'Quote_Time']
    _metrics = pd.DataFrame(index=_index, columns=['Value'])
    _straddlecall, _metrics.loc['Underlying', 'Value'], _metrics.loc['Quote_Time', 'Value'] =\
            price.get(optdata, 'call', midstrike, expiry1, False)
    _straddleput = price.get(optdata, 'put', midstrike, expiry1, False)[0]
    _farcall = price.get(optdata, 'call', highstrike, expiry2, False)[0]
    _farput = price.get(optdata, 'put', lowstrike, expiry2, False)[0]
    _metrics.loc['Straddle Call', 'Value'] = _straddlecall
    _metrics.loc['Straddle Put', 'Value'] = _straddleput
    _metrics.loc['Straddle Total', 'Value'] = _straddle_tot = _straddlecall + _straddleput
    _metrics.loc['Far Call', 'Value'] = _farcall
    _metrics.loc['Far Put', 'Value'] = _farput
    _metrics.loc['Far Total', 'Value'] = _far_tot = _farcall + _farput
    _metrics.loc['Straddle to Far Ratio', 'Value'] = _straddle_tot / _far_tot 
    _metrics.loc['Credit', 'Value'] = _straddle_tot - _far_tot
    return _metrics

def straddle(optdata, strike, expiry):
    """
    Metrics for evaluating a straddle

    Parameters
    ------------
    optdata : DataFrame
        Data returned from :func:`pynance.opt.retrieve.get`.
    strike : numeric
        Strike price.
    expiry : date or date str (e.g. '2015-01-01')
        Expiration date.

    Returns
    ------------
    metrics : DataFrame
        Metrics for evaluating straddle.
    """
    _callprice, _underlying, _qt = price.get(optdata, 'call', strike, expiry, False)
    _putprice = price.get(optdata, 'put', strike, expiry, False)[0]
    _index = ['Call', 'Put', 'Credit', 'Underlying', 'Quote_Time']
    _vals = np.array([_callprice, _putprice, _callprice + _putprice, _underlying, _qt])
    return pd.DataFrame(_vals, index=_index, columns=['Value'])



"""
.. Copyright (c) 2014, 2015 Marshall Farrier
   license http://opensource.org/licenses/MIT

Charts (:mod:`pynance.chart`)
========================================

.. currentmodule:: pynance.chart

Since detailed charts are readily available online, the charts
generated by this module serve 2 purposes:

1. Verify data integrity
2. Visualize events as defined by custom algorithms.

The functions in this submodule wrap pandas and matplotlib
for greater ease of use.

.. http://matplotlib.org/api/pyplot_api.html
   http://matplotlib.org/api/finance_api.html
   http://stackoverflow.com/questions/22027415/
   http://sentdex.com/sentiment-analysisbig-data-and-python-tutorials-algorithmic-trading/how-to-chart-stocks-and-forex-doing-your-own-financial-charting/
   http://stackoverflow.com/questions/19580116/plotting-candlestick-data-from-a-dataframe-in-python
   http://sentdex.com/sentiment-analysisbig-data-and-python-tutorials-algorithmic-trading/how-to-chart-stocks-and-forex-doing-your-own-financial-charting/
   http://pandas.pydata.org/pandas-docs/version/0.15.0/visualization.html
"""

import datetime as dt

import matplotlib.dates as mdates
import matplotlib.finance as fplt
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker

def candlestick(df, **kwargs):
    """
    Show and optionally save candlestick chart of a DataFrame as retrieved using
    :func:`pynance.data.get`.

    Parameters
    -----------
    df : DataFrame 
        Source data, containing columns 'Open', 'High', 'Low', 'Close' and 'Volume'.
    title : str, optional
        Title to be used for the chart.
    fname : str, optional
        If provided, the chart will be saved to a file named `fname`. `fname`
        should also include the extension '.png' or '.pdf'.
    events : DataFrame, optional
        Must have the same index as df.
        Up to 4 columns of events will be mapped. The order of the columns will
        determine the marker to be assigned to the event. The color order is:
        ``['g^', 'ro', 'bs', 'k^']``
        Non-events in this DataFrame should have a value like ``np.nan``,
        which matplotlib will not plot.
        The events DataFrame should use ``dtype='float'``. If not, a bug in numpy (or matplotlib)
        can lead to a :class:`TypeError`. Details
        `here <http://matplotlib.1069221.n5.nabble.com/type-error-with-python-3-2-and-version-1-1-1-of-matplotlib-numpy-error-td38784.html>`_.
    bollinger : DataFrame, optional
        If present, Bollinger bands will be overlaid. `bollinger` must have the same index as `df` and
        must contain columns 'Upper' and 'Lower'.
    sma : DataFrame, optional
        If present, first column will be overlaid as simple moving average. `sma`
        must have same index as `df`.

    Examples
    --------
    >>> pn.chart.candlestick(df, title='GE', fname='foo.png', events=evdf, eventcolors=['r', 'g'], 
            bollinger=bolldf, sma=smadf)
    """
    _make_chart(df, _candlestick_ax, **kwargs)

def adj_close(df, **kwargs):
    """
    Show and optionally save adj_close chart of a DataFrame as retrieved using :func:`pynance.data.get`.

    Examples
    --------
    >>> pn.chart.adj_close(df, title='GE', fname='foo.png', events=evdf, eventcolors=['r', 'g'], 
            bollinger=bolldf, sma=smadf)

    See also
    --------
    :func:`pynance.chart.candlestick`
    """
    _make_chart(df, _adj_close_ax, **kwargs)

def close(df, **kwargs):
    """
    Show and optionally save close chart of a DataFrame as retrieved using :func:`pynance.data.get`.

    Examples
    --------
    >>> pn.chart.close(df, title='GE', fname='foo.png', events=evdf, eventcolors=['r', 'g'], 
            bollinger=bolldf, sma=smadf)

    See also
    --------
    :func:`pynance.chart.candlestick`
    """
    _make_chart(df, _close_ax, **kwargs)

def _make_chart(df, chartfn, **kwargs):
    fig = plt.figure()
    ax1 = plt.subplot2grid((5, 4), (0, 0), rowspan=4, colspan=4)
    ax1.grid(True)
    plt.ylabel('Price')
    plt.setp(plt.gca().get_xticklabels(), visible=False)
    chartfn(df, ax1)
    if 'sma' in kwargs:
        _plot_sma(kwargs['sma'])
    if 'bollinger' in kwargs:
        _plot_bollinger(kwargs['bollinger'])
    if 'events' in kwargs:
        _plot_events(kwargs['events'])

    ax2 = plt.subplot2grid((5, 4), (4, 0), sharex=ax1, rowspan=1, colspan=4)
    ax2.bar(df.index, df.loc[:, 'Volume'])
    ax2.xaxis.set_major_locator(mticker.MaxNLocator(12))
    ax2.xaxis.set_major_formatter(mdates.DateFormatter('%Y-%m-%d'))
    if len(df.index) <= 128:
        # undesirable for longer time periods and eventually raises exception
        ax2.xaxis.set_minor_locator(mdates.DayLocator())
    ax2.yaxis.set_ticklabels([])
    ax2.grid(True)
    plt.ylabel('Volume')
    plt.xlabel('Date')
    plt.setp(plt.gca().get_xticklabels(), rotation=45, horizontalalignment='right')
    plt.subplots_adjust(left=.09, bottom=.18, right=.94, top=0.94, wspace=.20, hspace=0)
    if 'title' in kwargs:
        plt.suptitle(kwargs['title'])
    if 'fname' in kwargs:
        plt.savefig(kwargs['fname'], bbox_inches='tight')
    plt.show()
    plt.close()

def _candlestick_ax(df, ax):
    """
    # Alternatively: (but hard to get dates set up properly)
    plt.xticks(range(len(df.index)), df.index, rotation=45)
    fplt.candlestick2_ohlc(ax, df.loc[:, 'Open'].values, df.loc[:, 'High'].values, 
            df.loc[:, 'Low'].values, df.loc[:, 'Close'].values, width=0.2)
    """
    quotes = df.reset_index()
    quotes.loc[:, 'Date'] = mdates.date2num(quotes.loc[:, 'Date'].astype(dt.date))
    fplt.candlestick_ohlc(ax, quotes.values)

def _adj_close_ax(df, ax):
    ax.plot(df.index, df.loc[:, 'Adj Close'])

def _close_ax(df, ax):
    ax.plot(df.index, df.loc[:, 'Close'])

def _plot_bollinger(bolldf):
    plt.fill_between(bolldf.index, bolldf.loc[:, 'Upper'].values, 
            bolldf.loc[:, 'Lower'].values, facecolor='gray', alpha=0.5)

def _plot_sma(smadf):
    lines = plt.plot(smadf.index, smadf.iloc[:, 0].values)
    lines[0].set_color('g')

def _plot_events(events):
    colors = ['gd', 'rv', 'b*', 'k^']
    n_events = min({len(events.columns), len(colors)})
    for i in range(n_events):
        plt.plot(events.index, events.iloc[:, i].values, colors[i], alpha=0.5, ms=12.0)
